#include <setjmp.h>
#include "lanhep.h"

extern jmp_buf alg1_jmp_buf;

Term ProcInfsimal(Term t, Term ind)
	{
	Term t1;
	if(!is_compound(t) || CompoundArity(t)!=1)
		{
		ErrorInfo(701);
		puts("Wrong argument in 'infinitesimal' statement.");
		return 0;
		}
	t1=ConsumeCompoundArg(t,1);
	FreeAtomic(t);
	t=CommaToList(t1);
	for(t1=t;t1;t1=ListTail(t1))
		{
		int pw=1;
		Term aa=ListFirst(t1);
		if(is_compound(aa)&&CompoundName(aa)==OPR_DIV&&
				CompoundArg1(aa)==NewInteger(1))
		{
			aa=CompoundArg2(aa);
			pw=-1;
		}
		if(!is_atom(aa))
			{
			ErrorInfo(702);
			puts("Wrong argument in 'infinitesimal' statement.");
			}
		else
			{
			SetAtomProperty(aa,A_INFINITESIMAL,NewInteger(pw));
			if(pw==1)
				SetAtomProperty(ListFirst(t1),OPR_COEFF,NewInteger(1));
			}
		}
	FreeAtomic(t);		
	return 0;
	}

void alg1_rem_inf(Term a1)
	{
	List l, l1;
	l=ConsumeCompoundArg(a1,1);
	
	for(l1=l;l1;l1=l1?ListTail(l1):0)
		{
		Term m1;
		List l2;
		int  ino;
rep:
		ino=0;
		m1=ListFirst(l1);
		for(l2=CompoundArgN(m1,3);l2;l2=ListTail(l2))
			{
			Term prp=0;
			if(CompoundName(ListFirst(l2))==OPR_PARAMETER)
				prp=GetAtomProperty(CompoundArg2(ListFirst(l2)),A_INFINITESIMAL);
			if(prp && IntegerValue(prp)>0)
				ino+=IntegerValue(prp);
			if(CompoundName(ListFirst(l2))==OPR_WILD && 
					CompoundArg2(ListFirst(l2))==0)
				ino+=10000;
			}
		for(l2=CompoundArgN(m1,4);l2;l2=ListTail(l2))
			{
			Term prp=0;
			if(CompoundName(ListFirst(l2))==OPR_PARAMETER)
				prp=GetAtomProperty(CompoundArg2(ListFirst(l2)),A_INFINITESIMAL);
			if(prp && IntegerValue(prp)<0)
				ino++;
			}
		if(ino>1)
			{
			l2=ListTail(l1);
			l=CutFromList(l,l1);
			l1=l2;
			if(l1)
				goto rep;
			}
		else
		for(l2=CompoundArgN(m1,4);l2;l2=ListTail(l2))
			{
			Term prp=0;
			if(CompoundName(ListFirst(l2))==OPR_PARAMETER)
				prp=GetAtomProperty(CompoundArg2(ListFirst(l2)),A_INFINITESIMAL);
			if(prp && IntegerValue(prp)>0)
				{
				ErrorInfo(703);
				printf("infinitesimal parameter '%s' in denominator\n",
					AtomValue(CompoundArg2(ListFirst(l2))));
				longjmp(alg1_jmp_buf,1);
				}
			}
		}
	SetCompoundArg(a1,1,l);	
	}	
	
Term ProcTransform(Term t, Term ind)
	{
	Term t1;
	t1=ConsumeCompoundArg(t,1);
	FreeAtomic(t);

	if(is_compound(t1) && CompoundArity(t1)==2 && CompoundName(t1)==OPR_COMMA)
		{
		Term a1,a2;
		a1=ConsumeCompoundArg(t1,1);
		a2=ConsumeCompoundArg(t1,2);
		FreeAtomic(t1);
		ProcTransform(MakeCompound1(OPR_LET,a1),0);
		ProcTransform(MakeCompound1(OPR_LET,a2),0);
		return 0;
		}
		
	if(!is_compound(t1) || CompoundArity(t1)!=2 || CompoundName(t1)!=OPR_RARROW)
		{
		ErrorInfo(725);
		printf("bad argument in 'transform' statement\n");
		return 0;
		}

	ProcLet(MakeCompound1(OPR_LET,t1),0);
	
	return 0;
	}
	
static void alg1_dlt_to_wld(List ll, Label ind1, Label ind2, int d)
{
	List tind1, tind2;
	List l1;
	List r1,r2;
	Term ww;
	int i,j;
	
	l1=ConsumeCompoundArg(ListFirst(ll),1);
	FreeAtomic(ListFirst(ll));
	ChangeList(ll,0);
	
	tind1=ListFirst(l1);
	tind2=ListFirst(ListTail(l1));
	RemoveList(l1);
	
	
	r1=0;
	
	for(i=1;i<=d;i++)
	{
		Term mm, ww, aa;
		r2=0;
		
		for(j=1;j<=d;j++)
		{
			Term aa1;
			Term aa1ml=0;
			
			if(i==j)
			{	
				Term m1;
				m1=MakeCompound(A_MTERM,4);
				
				SetCompoundArg(m1,1,NewInteger(1));
				SetCompoundArg(m1,2,NewInteger(1));
				
				aa1ml=AppendLast(aa1ml,m1);
			}
			
			aa1=MakeCompound2(A_ALG1,aa1ml,0);
			r2=AppendLast(r2,aa1);
		}
		
		ww=MakeCompound2(OPR_WILD,MakeList1(CopyTerm(tind1)),r2);
		mm=MakeCompound(A_MTERM,4);
		SetCompoundArg(mm,1,NewInteger(1));
		SetCompoundArg(mm,2,NewInteger(1));
		SetCompoundArg(mm,3,MakeList1(ww));
		aa=MakeCompound2(A_ALG1,MakeList1(mm),MakeList1(CopyTerm(tind1)));
		r1=AppendLast(r1,aa);
		
	}
	
	ww=MakeCompound2(OPR_WILD,MakeList2(tind1,tind2),r1);
	ChangeList(ll,ww);
		
}
	


void alg1_fix_delta(Term a1)
{
	List l1,l2,il;
	
	il=NewList();
	for(l1=CompoundArg2(a1);l1;l1=ListTail(l1))
		il=AppendLast(il,CompoundArg2(ListFirst(l1)));
	
	
	for(l1=CompoundArg1(a1);l1;l1=ListTail(l1))
	{
		List ml;
		ml=ConsumeCompoundArg(ListFirst(l1),3);
	cnt:
		for(l2=ml;l2;l2=ListTail(l2))
		{
			if(CompoundArg2(ListFirst(l2))==A_DELTA)
			{
				Label in1, in2;
				int cno;
				
				in1=CompoundArg2(ListFirst(CompoundArg1(ListFirst(l2))));
				in2=CompoundArg2(ListNth(CompoundArg1(ListFirst(l2)),2));
				if(CompoundName(ListFirst(CompoundArg1(ListFirst(l2))))
																==OPR_WILD)
					cno=IntegerValue(CompoundArgN(
							ListFirst(CompoundArg1(ListFirst(l2))),3));
				else
					cno=0;
				
				if(in1==in2)
				{
					puts("Internal error (rmdlt1)");
					exit(0);
				}
				
				if(!cno && ListMember(il,in1) && ListMember(il,in2))
					continue;
				if(ListMember(il,in1) && ListMember(il,in2))
				{
					alg1_dlt_to_wld(l2,in1,in2,cno);
					continue;
				}
				
				if(ListMember(il,in1))
				{
					Label tmp;
					tmp=in1;
					in1=in2;
					in2=tmp;
				}
				
				ml=CutFromList(ml,l2);
				
				for(l2=ml;l2;l2=ListTail(l2))
				{
					List l3;
					for(l3=CompoundArg1(ListFirst(l2));l3;l3=ListTail(l3))
						if(CompoundArg2(ListFirst(l3))==in1)
						{
							SetCompoundArg(ListFirst(l3),2,in2);
							break;
						}
					if(l3)
						break;
				}
				
				if(!l2)
				{
					puts("Internal error (rmdlt2)");
					exit(0);
				}
				
				goto cnt;
			}
		}
		SetCompoundArg(ListFirst(l1),3,ml);
	}
	
	
}
