#include <stdio.h>
#include "lanhep.h"
#include <math.h>
#include <string.h>

static int dia_(int *, double *, double *, double *);

		
static int acmp(Term a1, Term a2)
	{
	return strcmp(AtomValue(a1),AtomValue(a2));
	}

static double mysqrt(double x)
{
	if(x<0)
		return -sqrt(-x);
	else
		return sqrt(x);
}
		
static int mcmp(Term m1, Term m2)
	{
	List l1,l2;
	l1=CompoundArg1(m1);
	l2=CompoundArg1(m2);
	if(ListLength(l1)>ListLength(l2))
		return 1;
	if(ListLength(l1)<ListLength(l2))
		return -1;
	return strcmp(AtomValue(ListFirst(l1)),
					AtomValue(ListFirst(l2)));
	}

static int prt_spin1(Atom p)
{
	Term prop;
	int spi;
	
	prop=GetAtomProperty(p,PROP_TYPE);
	if(CompoundName(prop)==OPR_FIELD)
		prop=GetAtomProperty(CompoundArg1(prop),PROP_TYPE);
	spi=IntegerValue(CompoundArgN(prop,4));
	if(spi==2 && CompoundArgN(prop,7)!=A_GAUGE)
		spi++;
	return spi;
}

	
static int prt_spin(Atom p)
	{
	Term prop;
	prop=GetAtomProperty(p,PROP_INDEX);
	if(is_empty_list(prop) || CompoundName(CompoundArg1(ListFirst(prop)))!=A_LORENTZ)
		return 0;
	prop=CompoundArg1(ListFirst(prop));
	if(CompoundArg1(prop)==NewInteger(2))
		return 2;
	return 1;
	}

	
static Atom remake_sp(List m3)
{
	Term mom1, mom2;
	List l;
	int ga=0;
	mom1=mom2=0;
	for(l=m3;l;l=ListTail(l))
	{
		if(CompoundName(ListFirst(l))==A_MOMENT)
		{
			if(mom1==0)
				mom1=ListFirst(l);
			else
				mom2=ListFirst(l);
		}
		if(CompoundName(ListFirst(l))==OPR_SPECIAL &&
				CompoundArg1(ListFirst(l))==A_GAMMA)
			ga++;
	}
	
	if(mom2==0 && ga)
		return NewAtom("G(p1)",0);
	if(mom2==0)
		return NewAtom("p1^mu",0);
	if(ListFirst(CompoundArg2(mom1))==ListFirst(CompoundArg2(mom2)))
		return NewAtom("p1**2",0);
	return NewAtom("p1^mu*p1^nu",0);
}


static void wrt_ml(FILE *f, List ml)
	{
	List gl;
	List l;
	gl=NewList();
	for(l=ml;l;l=ListTail(l))
		{
		Term sp;
		if(is_empty_list(CompoundArgN(ListFirst(l),3)))
			sp=0;
		else
			sp=CompoundArg1(ListFirst(CompoundArgN(ListFirst(l),3)));
		if(is_integer(sp))
			sp=remake_sp(CompoundArgN(ListFirst(l),3));
		if(!ListMember(gl,sp))
			gl=AppendLast(gl,sp);
		}
	
	for(l=gl;l;l=ListTail(l))
		{
		List l1;
		fprintf(f,"(");
		for(l1=ml;l1;l1=ListTail(l1))
			{
			Term sp;
			List l2;
			if(is_empty_list(CompoundArgN(ListFirst(l1),3)))
				sp=0;
			else
				sp=CompoundArg1(ListFirst(CompoundArgN(ListFirst(l1),3)));
			if(is_integer(sp))
				sp=remake_sp(CompoundArgN(ListFirst(l1),3));
			if(ListFirst(l)!=sp)
				continue;
			{
			long n,d;
			n=IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l1))));
			d=IntegerValue(CompoundArg2(CompoundArg1(ListFirst(l1))));
			fprintf(f,"%+ld",n);
			if(d!=1)
				fprintf(f,"/%ld",d);
			}
			for(l2=CompoundArg2(ListFirst(l1));l2;l2=ListTail(l2))
				{
				Atom p;
				int pw;
				p=CompoundArg1(ListFirst(l2));
				pw=IntegerValue(CompoundArg2(ListFirst(l2)));
				if(pw>0)
					fprintf(f,"*%s",AtomValue(p));
				else
					fprintf(f,"/%s",AtomValue(p));
				if(pw<0)
					pw=-pw;
				if(pw>1)
					fprintf(f,"**%d",pw);
				}
			}
			
		fprintf(f,")");
		if(ListFirst(l) && ListFirst(l)!=A_DELTA)
			{
			if(is_atom(ListFirst(l)))
				fprintf(f,"*%s",AtomValue(ListFirst(l)));
			else
				fprintf(f,"*p%ld",IntegerValue(ListFirst(l)));
			}
		if(ListTail(l))
			fprintf(f,"+");
		}
	FreeAtomic(gl);
	}

static void wrt_num(FILE *f, List ml, int spi)
	{
	List gl;
	List l;
	gl=NewList();
	for(l=ml;l;l=ListTail(l))
		{
		Term sp;
		if(is_empty_list(CompoundArgN(ListFirst(l),3)))
			sp=0;
		else
			sp=CompoundArg1(ListFirst(CompoundArgN(ListFirst(l),3)));
		if(!ListMember(gl,sp))
			gl=AppendLast(gl,sp);
		}
	
	for(l=gl;l;l=ListTail(l))
		{
		List l1;
		double vi,vr;
		int ni,nr;
		vi=0.0;
		vr=0.0;
		ni=0;
		nr=0;
		
		for(l1=ml;l1;l1=ListTail(l1))
			{
			Term sp;
			double v1;
			List l2;
			int is_i=0;
			if(is_empty_list(CompoundArgN(ListFirst(l1),3)))
				sp=0;
			else
				sp=CompoundArg1(ListFirst(CompoundArgN(ListFirst(l1),3)));
			if(ListFirst(l)!=sp)
				continue;
			v1=(double)IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l1))));
			v1/=(double)IntegerValue(CompoundArg2(CompoundArg1(ListFirst(l1))));
			if(CompoundArg2(ListFirst(l1)))
				{
				v1*=EvalParameter(CompoundArg2(ListFirst(l1)));
				for(l2=CompoundArg2(ListFirst(l1));l2;l2=ListTail(l2))
					{
					if(CompoundArg1(ListFirst(l2))==A_I)
						is_i=1;
					}
				}
			else
				nr++;
			if(is_i)
				{
				vi+=v1;
				ni++;
				}
			else
				{
				vr+=v1;
				nr++;
				}
			}
		if(nr && ni)
			fprintf(f,"(%+f%+f*i)",vr,vi);
		else
			{
			if(nr)
			{
				if(spi==1)
					fprintf(f,"%+f",vr);
				else
					fprintf(f,"%+f**2",mysqrt(vr));
			}
			else
				fprintf(f,"%+f*i",vi);
			}
		if(ListFirst(l) && ListFirst(l)!=A_DELTA)
			{
			if(is_atom(ListFirst(l)))
				fprintf(f,"*%s",AtomValue(ListFirst(l)));
			else
				fprintf(f,"*p%ld",IntegerValue(ListFirst(l)));
			}
		}
	FreeAtomic(gl);
	}
	

static List prt_with_kt=0;
static List prt_with_mt=0;
static int correct_kt(Term m2, int spi);
static int correct_ktv(List m2);
static int correct_mt(Term t, Atom bn, Atom *m);

	
static void check_momenta(Term a2, Atom b1, Atom b2, FILE *f)
{
	List l1,l2,l3,lm;
	
	l1=ConsumeCompoundArg(a2,4);
rpt1:
	for(l2=l1;l2;l2=ListTail(l2))
	{
		for(l3=CompoundArgN(ListFirst(l2),3);l3;l3=ListTail(l3))
		{
			if(CompoundName(ListFirst(l3))==A_MOMENT)
			{
				l1=CutFromList(l1,l2);
				goto rpt1;
			}
		}
	}
	SetCompoundArg(a2,4,l1);
	lm=NewList();
	l1=ConsumeCompoundArg(a2,5);
rpt2:
	for(l2=l1;l2;l2=ListTail(l2))
	{
		for(l3=CompoundArgN(ListFirst(l2),3);l3;l3=ListTail(l3))
		{
			if(CompoundName(ListFirst(l3))==A_MOMENT)
			{
				lm=AppendLast(lm,ListFirst(l2));
				ChangeList(l2,0);
				l1=CutFromList(l1,l2);
				goto rpt2;
			}
		}
	}
	SetCompoundArg(a2,5,l1);
	if(is_empty_list(lm))
		return;
/*	
	p1=CompoundArg1(ListFirst(CompoundArg1(a2)));
	p2=CompoundArg1(ListFirst(ListTail(CompoundArg1(a2))));
*/	
	
/*	WriteTerm(CompoundArg1(a2));printf(" base ");
	WriteTerm(b1);printf("/");WriteTerm(b2);printf("; sp=%d\n",prt_spin(b1));
	puts("");
*/
	
	if(b1!=b2)
	{
		fprintf(f,"\n----------------------------------------------------------------\n");
		fprintf(f,"Mixing of particles ");
		fWriteTerm(f,CompoundArg1(a2));
		fprintf(f," in kinetic term:\n");
		wrt_ml(f,lm);
		FreeAtomic(lm);
		return;
	}
	
	prt_with_kt=AppendLast(prt_with_kt,b1);
	
	if(ListLength(lm)==2 && prt_spin1(b1)==3 && correct_ktv(lm))
		return;
	
	if(ListLength(lm)>1 || !correct_kt(ListFirst(lm),prt_spin(b1)))
	{
		fprintf(f,"\n----------------------------------------------------------------\n");
		fprintf(f,"Incorrect kinetic term for particle ");
		fWriteTerm(f,CompoundArg1(a2));
		fprintf(f," :\n");
		wrt_ml(f,lm);
		FreeAtomic(lm);
		return;	
	}
	
}
	
static void look_for_orth(List prt, List al, int c, int s, FILE *f);

extern Label orth_del_offdiag;
extern Label orth_red_diag;
static int do_modify=0;

Term ProcCheckMasses(Term t, Term ind)
	{
	
	List l, all_lagr;
	FILE *outf;
	List alone_prt=0, ds_prt=0;
	List masses=0;
	Atom p1, p2, b1, b2;
	
	do_modify=0;
	
	if(is_compound(t))
	{
		int i;
		for(i=1;i<=CompoundArity(t);i++)
		{
			Term a;
			a=CompoundArgN(t,i);
			if(is_atom(a) && strcmp(AtomValue(a),"modify")==0)
			{
				do_modify=1;
				continue;
			}
			if(is_compound(a) && CompoundName(a)==OPR_EQSIGN &&
					CompoundArity(a)==2 &&
					CompoundArg1(a)==NewAtom("DelOffDiag",0) &&
					is_atom(CompoundArg2(a)))
			{
				Atom v;
				Term prop;
				v=CompoundArg2(a);
				prop=GetAtomProperty(v,A_ORTH_MATR);
				if(!is_compound(prop))
				{
					ErrorInfo(703);
					printf("CheckMasses: '%s' is not a matrix element.\n",
							AtomValue(v));
					continue;
				}
				orth_del_offdiag=CompoundArg1(prop);
				continue;
			}
			if(is_compound(a) && CompoundName(a)==OPR_EQSIGN &&
					CompoundArity(a)==2 &&
					CompoundArg1(a)==NewAtom("RedDiag",0) &&
					is_atom(CompoundArg2(a)))
			{
				Atom v;
				Term prop;
				v=CompoundArg2(a);
				prop=GetAtomProperty(v,A_ORTH_MATR);
				if(!is_compound(prop))
				{
					ErrorInfo(703);
					printf("CheckMasses: '%s' is not a matrix element.\n",
							AtomValue(v));
					continue;
				}
				orth_red_diag=CompoundArg1(prop);
				continue;
			}
			ErrorInfo(709);
			printf("CheckMasses: invalid argument '");WriteTerm(a);puts("'.");
			return 0;
		}
	}
				
	
	outf=fopen("masses.chk","w");
	
	
	all_lagr=all_vert_list();
	
	for(l=all_lagr;l;l=ListTail(l))
		{
		Term a2;
		Term a2_so;
		Term prtl;
		List l1, l2;
		
		a2=ListFirst(l);
		prtl=CompoundArg1(a2);
		
				
		if(ListLength(prtl)>2 || is_empty_list(prtl))
			continue;
			
		a2=CopyTerm(a2);
		alg2_symmetrize(a2);
		alg2_common_s(a2);
		alg2_common_n(a2);
/*		alg2_common_t(a2);
		FreeAtomic(ConsumeCompoundArg(a2,4));
*/
		alg2_red_cos(a2);

		a2_so=CopyTerm(a2);
		alg2_red_orth(a2);

		alg2_red_sico(a2);
		alg2_red_sico(a2_so);

		alg2_red_comsico(a2);
		alg2_red_comsico(a2_so);
		
		alg2_recommon_n(a2);
		alg2_recommon_n(a2_so);
		
		alg2_recommon_s(a2);
		alg2_recommon_s(a2_so);
		
		alg2_red_1pm5(a2);
		alg2_red_1pm5(a2_so);
		
		if(is_empty_list(CompoundArgN(a2,5)))
			{
			FreeAtomic(a2);
			FreeAtomic(a2_so);
			continue;
			}
		
		alg2_decommon_s(a2);
		alg2_decommon_s(a2_so);
		
		alg2_decommon_n(a2);
		alg2_decommon_n(a2_so);

		
		if(ListLength(prtl)==1)
			{
			Atom p;
			prtl=ConsumeCompoundArg(a2,1);
			p=CompoundArg1(ListFirst(prtl));
			FreeAtomic(prtl);
			SetCompoundArg(a2,1,p);
			alone_prt=AppendFirst(alone_prt,a2);
			FreeAtomic(a2_so);
			continue;
			}
		
		SetCompoundArg(a2,4,ConsumeCompoundArg(a2_so,5));
		FreeAtomic(a2_so);
		
			
		prtl=ConsumeCompoundArg(a2,1);
		p1=CompoundArg1(ListFirst(prtl));
		p2=CompoundArg1(ListFirst(ListTail(prtl)));
		FreeAtomic(prtl);
		
		SetCompoundArg(a2,1,MakeCompound2(OPR_DIV,p1,p2));

		if(prt_spin(p1)!=prt_spin(p2))
			{
			ds_prt=AppendLast(ds_prt,a2);
			continue;
			}
		
		
		prtl=GetAtomProperty(p1,PROP_TYPE);
		if(CompoundName(prtl)==OPR_FIELD && CompoundArg2(prtl)==NewInteger(4))
			p1=CompoundArg1(prtl);
			
		prtl=GetAtomProperty(p2,PROP_TYPE);
		if(CompoundName(prtl)==OPR_FIELD && CompoundArg2(prtl)==NewInteger(4))
			p2=CompoundArg1(prtl);
			
		
		prtl=GetAtomProperty(p1,PROP_TYPE);
		if(CompoundName(prtl)==OPR_FIELD)
			{
			Atom bp;
			bp=CompoundArg1(prtl);
			prtl=GetAtomProperty(bp,PROP_TYPE);
			if(bp!=CompoundArg1(prtl))
				b1=GetAtomProperty(p1,A_ANTI);
			else
				b1=p1;
			}
		else
			b1=CompoundArg1(prtl);
		
		prtl=GetAtomProperty(p2,PROP_TYPE);
		if(CompoundName(prtl)==OPR_FIELD)
			{
			Atom bp;
			bp=CompoundArg1(prtl);
			prtl=GetAtomProperty(bp,PROP_TYPE);
			if(bp!=CompoundArg1(prtl))
				b2=GetAtomProperty(p2,A_ANTI);
			else
				b2=p2;
			}
		else
			b2=CompoundArg1(prtl);
		
		SetCompoundArg(a2,1,MakeCompound2(OPR_DIV,p1,p2));
		
		check_momenta(a2,b1,b2,outf);
		
		if(prt_spin(b1)<2)
		{
			for(l1=CompoundArgN(a2,5);l1;l1=ListTail(l1))
			{
				int n;
				n=IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l1))));
				n=-n;
				SetCompoundArg(CompoundArg1(ListFirst(l1)),1,NewInteger(n));
			}
			
			for(l1=CompoundArgN(a2,4);l1;l1=ListTail(l1))
			{
				int n;
				n=IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l1))));
				n=-n;
				SetCompoundArg(CompoundArg1(ListFirst(l1)),1,NewInteger(n));
			}
		}
		
		if(is_empty_list(CompoundArgN(a2,5)))
		{
			FreeAtomic(a2);
			continue;
		}
		
		if(!ListMember(prt_with_mt,b1))
			prt_with_mt=AppendLast(prt_with_mt,b1);
		if(!ListMember(prt_with_mt,b2))
			prt_with_mt=AppendLast(prt_with_mt,b2);
		
		
		for(l1=masses;l1;l1=ListTail(l1))
			{
			Term t;
			t=ListFirst(l1);

			if(ListMember(CompoundArg1(t),b1) || ListMember(CompoundArg1(t),b2))
				{
				if(!ListMember(CompoundArg1(t),b1))
					{
					l2=ConsumeCompoundArg(t,1);
					l2=AppendLast(l2,b1);
					SetCompoundArg(t,1,l2);
					}
				if(!ListMember(CompoundArg1(t),b2))
					{
					l2=ConsumeCompoundArg(t,1);
					l2=AppendLast(l2,b2);
					SetCompoundArg(t,1,l2);
					}

				l2=ConsumeCompoundArg(t,2);
				l2=AppendLast(l2,a2);
				SetCompoundArg(t,2,l2);
				break;
				}
			}
			
		if(is_empty_list(l1))
			masses=AppendLast(masses,MakeCompound2(OPR_DIV,
				b1==b2?MakeList1(b1):MakeList2(b1,b2),
				MakeList1(a2)));
		}
	
		
	FreeAtomic(all_lagr);

	/* end of all_vrt processing */
	
	{
		List l1,l2;
		for(l1=masses;l1;l1=ListTail(l1))
		{
		rpt:
			for(l2=ListTail(l1);l2;l2=ListTail(l2))
			{
				List l3;
				int ff=0;
				for(l3=CompoundArg1(ListFirst(l1));l3;l3=ListTail(l3))
					if(ListMember(CompoundArg1(ListFirst(l2)),ListFirst(l3)))
						ff=1;
				for(l3=CompoundArg1(ListFirst(l2));l3;l3=ListTail(l3))
					if(ListMember(CompoundArg1(ListFirst(l1)),ListFirst(l3)))
						ff=1;
				if(ff)
				{
					List lp,la;
					lp=ConsumeCompoundArg(ListFirst(l2),1);
					la=ConsumeCompoundArg(ListFirst(l2),2);
					CutFromList(masses,l2);
					l2=ConsumeCompoundArg(ListFirst(l1),1);
					for(l3=lp;l3;l3=ListTail(l3))
						if(!ListMember(l2,ListFirst(l3)))
							l2=AppendLast(l2,ListFirst(l3));
					SetCompoundArg(ListFirst(l1),1,l2);
					RemoveList(lp);
					l2=ConsumeCompoundArg(ListFirst(l1),2);
					l2=ConcatList(l2,la);
					SetCompoundArg(ListFirst(l1),2,l2);
					goto rpt;
				}
			}
		}
	}
	
	
	{
		List l1,l2;
		l2=NewList();
		for(l1=all_prtc_list();l1;l1=ListTail(l1))
		{
			Atom p;
			Term prop;
			p=ListFirst(l1);
			prop=GetAtomProperty(p,PROP_TYPE);
			if(CompoundName(prop)==OPR_FIELD)
				continue;
			if(p!=CompoundArg1(prop))
				continue;
			if(CompoundArgN(prop,7)==OPR_MLT)
				continue;
			if(!ListMember(prt_with_kt,ListFirst(l1)))
				l2=AppendLast(l2,ListFirst(l1));
		}
		if(l2)
		{
			fprintf(outf,"\n----------------------------------------------------------------\n");
			fprintf(outf,"Particles without kinetic term:\n");
			fWriteTerm(outf,l2);
			fprintf(outf,"\n");
			FreeAtomic(l2);
		}
	}
	
	{
		List l1,l2;
		l2=NewList();
		for(l1=all_prtc_list();l1;l1=ListTail(l1))
		{
			Atom p;
			Term prop;
			p=ListFirst(l1);
			prop=GetAtomProperty(p,PROP_TYPE);
			if(CompoundName(prop)==OPR_FIELD)
				continue;
			if(p!=CompoundArg1(prop))
				continue;
			if(CompoundArgN(prop,7)==OPR_MLT)
				continue;
			if(CompoundArgN(prop,5)==0)
				continue;
			if(!ListMember(prt_with_mt,ListFirst(l1)))
				l2=AppendLast(l2,ListFirst(l1));
		}
		if(l2)
		{
			fprintf(outf,"\n----------------------------------------------------------------\n");
			fprintf(outf,"Particles without mass term:\n");
			fWriteTerm(outf,l2);
			fprintf(outf,"\n");
			FreeAtomic(l2);
		}
	}
	
		
	for(l=masses;l;l=ListTail(l))
		if(ListLength(CompoundArg1(ListFirst(l)))>1)
			{
			List l1;
			l1=ConsumeCompoundArg(ListFirst(l),1);
			l1=SortedList(l1,acmp);
			SetCompoundArg(ListFirst(l),1,l1);
			}
			
	masses=SortedList(masses,mcmp);
		
	
	if(alone_prt)
		fprintf(outf,"\n----------------------------------------------------------------\n");
	for(l=alone_prt;l;l=ListTail(l))
		{
		fprintf(outf,"Linear term with particle %s\n",AtomValue(CompoundArg1(ListFirst(l))));
		fprintf(outf,"Expression: ");
		wrt_ml(outf,CompoundArgN(ListFirst(l),5));
		fprintf(outf,"\nNumeric value: ");
		wrt_num(outf,CompoundArgN(ListFirst(l),5),1);
		fprintf(outf,"\n\n");
		}
		
	if(ds_prt)
		fprintf(outf,"\n----------------------------------------------------------------\n");
	for(l=ds_prt;l;l=ListTail(l))
		{
		fprintf(outf,"Mixing particles with different spin: ");
		fWriteTerm(outf,CompoundArg1(ListFirst(l)));
		fprintf(outf,"\n");
		fprintf(outf,"Expression: ");
		wrt_ml(outf,CompoundArgN(ListFirst(l),5));
		fprintf(outf,"\nNumeric value: ");
		wrt_num(outf,CompoundArgN(ListFirst(l),5),1);
		fprintf(outf,"\n\n");
		}
	
	for(l=masses;l;l=ListTail(l))
		{
		int is_c, spi;
		List l1,l2,al;
		int i1,i2; 
		
		l1=CompoundArg1(ListFirst(l));
		if(ListLength(l1)==1)
		{
			Atom tmass;
			Term a2;
			Atom bn;
			bn=ListFirst(l1);
			a2=ListFirst(CompoundArg2(ListFirst(l)));
			
			if(!correct_mt(a2,bn,&tmass))
			{
				fprintf(outf,"\n---------------------------------------------------------------\n");
				fprintf(outf,"Mass term for particle ");
				fWriteTerm(outf,bn);
				fprintf(outf," :\n");
				wrt_ml(outf,CompoundArgN(a2,5));
				fprintf(outf," = ");
				wrt_num(outf,CompoundArgN(a2,5),prt_spin(bn));
				fprintf(outf,"\n");
				if(is_atom(tmass))
				{
					fprintf(outf,"This particles was declared with mass ");
					fWriteTerm(outf,tmass);
					fprintf(outf," = %f\n",EvalParameter(tmass));
				}
				else
				{
					if(tmass==0)
						fprintf(outf,"This particles was declared massless\n");
					else
						fprintf(outf,"This (derived) particle should not have mass term\n");
				}
			}
				
			continue;
		}
		
		fprintf(outf,"\n---------------------------------------------------------------\n");
		fprintf(outf,"Mixing of particles ");
		fWriteTerm(outf,l1);
		fprintf(outf,"\n");
		
		is_c=(ListFirst(l1)!=GetAtomProperty(ListFirst(l1),A_ANTI));
		spi=prt_spin(ListFirst(l1));
		for(l2=l1;l2;l2=ListTail(l2))
			{
			int is_c1;
			is_c1=(ListFirst(l2)!=GetAtomProperty(ListFirst(l2),A_ANTI));
			if(is_c1 != is_c)
				{
				is_c=1;
				fprintf(outf,"Warning: charged and uncharged particles mixed\n");
				break;
				}
			}
			
		al=ConsumeCompoundArg(ListFirst(l),2);
		
		for(i1=1;i1<=ListLength(l1);i1++)
		for(i2=1;i2<=ListLength(l1);i2++)
			{
			if(is_c==0 && i1>i2)
				continue;
				
			p1=GetAtomProperty(ListNth(l1,i1),A_ANTI);
			p2=ListNth(l1,i2);
			if(spi!=1 && (strcmp(AtomValue(p1),AtomValue(p2))>0))
				{
				Atom tmp;
				tmp=p1;
				p1=p2;
				p2=tmp;
				}
			fprintf(outf,"\nM%d%d (%s/%s) = ",i1,i2,AtomValue(p1),AtomValue(p2));
			for(l2=al;l2;l2=ListTail(l2))
				{
				if(CompoundArg1(CompoundArg1(ListFirst(l2)))==p1 &&
					CompoundArg2(CompoundArg1(ListFirst(l2)))==p2)
					{
					wrt_ml(outf,CompoundArgN(ListFirst(l2),5));
/*					fprintf(outf,"\n    = ");
					wrt_ml(outf,CompoundArgN(ListFirst(l2),4));*/
					fprintf(outf,"\n    = ");
					wrt_num(outf,CompoundArgN(ListFirst(l2),5),spi);
					fprintf(outf,"\n");

					break;
					}
				}
			if(is_empty_list(l2))
				{
				fprintf(outf,"0\n");
				}
				
			if(i1==i2)
				{
				Term ms=0,p;
				p=ListNth(l1,i1);
				p=GetAtomProperty(p,PROP_TYPE);
				if(p && CompoundName(p)==OPR_PARTICLE)
					ms=CompoundArgN(p,5);
				if(ms)
					fprintf(outf,"    Declared mass %s=%f\n",AtomValue(ms),
							EvalParameter(ms));
				}
			fprintf(outf,"\n");	
			}
			
		fprintf(outf,"\n");
		
		look_for_orth(l1,al,is_c,spi,outf);
			
		}
		
		
	fclose(outf);
	puts("File 'masses.chk' is created"); 
	return 0;
	}

static List get_expr(List ml, int i, int j, int fixel, int is_c)
{
	List l1,l2,l3;
	List res=0;
	
	
	if(i>j)
	{
		int tmp;
		tmp=i;
		i=j;
		j=tmp;
	}
	
	for(l1=ml;l1;l1=ListTail(l1))
	{
		int ti,tj;
		List p1=0,p2=0;
		Term prop;
		for(l3=CompoundArg2(ListFirst(l1));l3;l3=ListTail(l3))
		{
			if((prop=GetAtomProperty(CompoundArg1(ListFirst(l3)),A_ORTH_MATR))&&
					CompoundArg1(prop)!=orth_red_diag)
			{
				if(p1==0)
					p1=l3;
				else
					p2=l3;
			}
		}
		
		prop=GetAtomProperty(CompoundArg1(ListFirst(p1)),A_ORTH_MATR);
		ti=IntegerValue(CompoundArgN(prop,5-fixel));
		if(p2==0)
			tj=ti;
		else
		{
			prop=GetAtomProperty(CompoundArg1(ListFirst(p2)),A_ORTH_MATR);
			tj=IntegerValue(CompoundArgN(prop,5-fixel));
		}
		
		if(tj<ti)
		{
			int tmp;
			tmp=ti;
			ti=tj;
			tj=tmp;
		}
		
		if(ti!=i || tj!=j)
		{
			/*printf("found %d %d; cont...\n",ti,tj);*/
			continue;
		}
		
		p1=CopyTerm(ListFirst(l1));
		
		l2=ConsumeCompoundArg(p1,2);

rpt:	for(l3=l2;l3;l3=ListTail(l3))
		{
			Term pp;
			if((pp=GetAtomProperty(CompoundArg1(ListFirst(l3)),A_ORTH_MATR))&&
					CompoundArg1(pp)!=orth_red_diag)
			{
				l2=CutFromList(l2,l3);
				goto rpt;
			}
		}
		SetCompoundArg(p1,2,l2);
		for(l2=res;l2;l2=ListTail(l2))
			if(EqualTerms(CompoundArg2(p1),CompoundArg2(ListFirst(l2))) &&
			   EqualTerms(CompoundArgN(p1,3),CompoundArgN(ListFirst(l2),3)))
				break;
		if(l2)
		{
			int n,d,n1,d1,n2,d2,cf;
			n1=IntegerValue(CompoundArg1(CompoundArg1(p1)));
			d1=IntegerValue(CompoundArg2(CompoundArg1(p1)));
			n2=IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l2))));
			d2=IntegerValue(CompoundArg2(CompoundArg1(ListFirst(l2))));
			n=n1*d2+n2*d1;
			d=d1*d2;
			cf=gcf(n,d);
			n/=cf;
			d/=cf;
			SetCompoundArg(CompoundArg1(ListFirst(l2)),1,NewInteger(n));
			SetCompoundArg(CompoundArg1(ListFirst(l2)),2,NewInteger(d));
			FreeAtomic(p1);
		}
		else
			res=AppendLast(res,p1);
	}
	
	for(l2=res;l2;l2=ListTail(l2))
	{
		int n,d,cf;
		n=IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l2))));
		d=IntegerValue(CompoundArg2(CompoundArg1(ListFirst(l2))));
/*		if(is_c==0)
			d*=2;*/
		if(i!=j)
			d*=2;
		cf=gcf(n,d);
		n/=cf;
		d/=cf;
		SetCompoundArg(CompoundArg1(ListFirst(l2)),1,NewInteger(n));
		SetCompoundArg(CompoundArg1(ListFirst(l2)),2,NewInteger(d));
	}
	/*printf("%d %d fixel=%d ",i,j,fixel);WriteTerm(res);puts("");*/
	return res;
}
		
static void look_for_orth(List prt, List al, int is_c, int spi, FILE *f)
{
	List l1,l2;
	Label ola;
	int prtlen, fixel=0, i, j;
	
	Atom prta[32];
	int  prtp[32];
	Term mmatr[32][32];
	
	double smatr[32][32], smatr1[32][32], zmatr[32][32], rmatr[32];
/*
fprintf(f,"look for orth: ");fWriteTerm(f,prt);fprintf(f,"\n");
*/		
	ola=0;
	for(l1=al;l1;l1=ListTail(l1))
		for(l2=CompoundArgN(ListFirst(l1),4);l2;l2=ListTail(l2))
		{
			int macoe=0;
			List l3;
			for(l3=CompoundArg2(ListFirst(l2));l3;l3=ListTail(l3))
			{
				Term prop;
				prop=GetAtomProperty(CompoundArg1(ListFirst(l3)),A_ORTH_MATR);
				if(!prop)
					continue;
				if(CompoundArg1(prop)==orth_red_diag)
					continue;
				if(ola && ola!=CompoundArg1(prop))
					{/*fprintf(f,"lfo: mtr!=:");
					fWriteTerm(f,CompoundArg1(ListFirst(l3)));
					fprintf(f,"\n");*/ return;}
				/*if(ola==0)
				{	fprintf(f,"lfo: mtr==:");
					fWriteTerm(f,CompoundArg1(ListFirst(l3)));
					fprintf(f,"\n");
				}	*/
				ola=CompoundArg1(prop);
				macoe+=IntegerValue(CompoundArg2(ListFirst(l3)));
			}
			if(macoe!=2)
				{/*fprintf(f,"lfo: macoe!=2 \n");*/ return;}
		}
	


	prtlen=ListLength(prt);
	if(is_c==0 && 2*ListLength(al)!=prtlen*(prtlen+1))
		return;
	if(is_c!=0 &&   ListLength(al)!=prtlen*prtlen)
		return;


	for(i=1;i<=prtlen;i++)
	{
		prta[i]=ListNth(prt,i);
		prtp[i]=0;
	}
	
	for(i=1;i<=prtlen;i++)
	{
		Atom p1,p2;
		p2=prta[i];
		p1=GetAtomProperty(p2,A_ANTI);
		if(spi!=1 && (strcmp(AtomValue(p1),AtomValue(p2))>0))
				{
				Atom tmp;
				tmp=p1;
				p1=p2;
				p2=tmp;
				}
		for(l1=al;l1;l1=ListTail(l1))
		{
			if(CompoundArg1(CompoundArg1(ListFirst(l1)))==p1 &&
				CompoundArg2(CompoundArg1(ListFirst(l1)))==p2)
			{
				List oeused=0, e1used=0, e2used=0;
				List l3;
						
				for(l2=CompoundArgN(ListFirst(l1),4);l2;l2=ListTail(l2))
					for(l3=CompoundArg2(ListFirst(l2));l3;l3=ListTail(l3))
					{
						Atom p;
						Term prp1;
						p=CompoundArg1(ListFirst(l3));
						if((prp1=GetAtomProperty(p,A_ORTH_MATR)) && 
								CompoundArg1(prp1)!=orth_red_diag &&
								!ListMember(oeused,p))
							oeused=AppendLast(oeused,p);
					}
					
				for(l2=oeused;l2;l2=ListTail(l2))
				{
					Term prop;
					Integer i,j;
					prop=GetAtomProperty(ListFirst(l2),A_ORTH_MATR);
					i=CompoundArgN(prop,3);
					j=CompoundArgN(prop,4);
					if(!ListMember(e1used,i)) e1used=AppendLast(e1used,i);
					if(!ListMember(e2used,j)) e2used=AppendLast(e2used,j);
				}


				if((ListLength(e1used)==1&&ListLength(e2used)==1) ||
						(ListLength(e1used)!=1&&ListLength(e2used)!=1))
					return;
				
				if(ListLength(e1used)==1)
				{
					if(fixel==2) return;
					fixel=1;
					prtp[i]=IntegerValue(ListFirst(e1used));
				}
				else
				{
					if(fixel==1) return;
					fixel=2;
					prtp[i]=IntegerValue(ListFirst(e2used));
				}
				
				
				FreeAtomic(e1used);
				FreeAtomic(e2used);
				FreeAtomic(oeused);
				break;
			}
		}
		
	}
	
	
	for(l1=al;l1;l1=ListTail(l1))
	{
		for(i=1;i<=prtlen;i++)
		for(j=i;j<=prtlen;j++)
		{
			Term thexpr;
			thexpr=get_expr(CompoundArgN(ListFirst(l1),4),
					prtp[i],prtp[j],fixel,is_c);
			if(l1==al)
				mmatr[i][j]=thexpr;
			else
			{
				if(!EqualTerms(thexpr,mmatr[i][j]))
				{
					/*
					printf("redef %d %d ",i,j);				
					WriteTerm(CompoundArg1(ListFirst(l1)));
					printf(" : ");
					wrt_ml(stdout,thexpr);
					puts("");
					*/
					return;
				}
				FreeAtomic(thexpr);
			}
		}
	}
	
	fprintf(f,"It is recognized that these fields are rotated by matrix:\n\n");
	
	for(i=1;i<=prtlen;i++)
	{
		fprintf(f,"  ( ");
		for(j=1;j<=prtlen;j++)
			fprintf(f," %s ",AtomValue(herm_matr_el(ola,prtp[i],prtp[j])));
		fprintf(f," )\n");
	}
	
	fprintf(f,"\nThe mass matrix before introducing this rotation:\n\n");
	
	for(i=1;i<=prtlen;i++)
	for(j=i;j<=prtlen;j++)
	{
		fprintf(f,"M%d%d = ",i,j);
		if(mmatr[i][j])
			wrt_ml(f,mmatr[i][j]);
		else
			fprintf(f,"0");
		fprintf(f,"\n");

	}
	
	
	
	for(i=1;i<=prtlen;i++)
	for(j=i;j<=prtlen;j++)
	{
		double tv;
		
		if(mmatr[i][j]==0)
		{
			smatr[i-1][j-1]=0.0;
			smatr[j-1][i-1]=0.0;
			continue;
		}
		
		tv=0.0;
		for(l1=mmatr[i][j];l1;l1=ListTail(l1))
		{
			double ttv;
			ttv=(double)IntegerValue(CompoundArg1(CompoundArg1(ListFirst(l1))));
			ttv/=(double)IntegerValue(CompoundArg2(CompoundArg1(ListFirst(l1))));
			for(l2=CompoundArg2(ListFirst(l1));l2;l2=ListTail(l2))
			{
				if(CompoundArg1(ListFirst(l2))==A_I)
					return;
				ttv*=EvalParameter(ListFirst(l2));
			}
			tv+=ttv;
		}
		
		if(is_c==0 && spi==1 && i==j)
		{
			smatr[i-1][j-1]=tv;
			smatr[j-1][i-1]=tv;
		}
		else
		{
			smatr[i-1][j-1]=tv;
			smatr[j-1][i-1]=tv;
		}
	}
	/*
	for(i=1;i<=prtlen;i++)
	{
	fprintf(f,"s %dx ",i);
	for(j=1;j<=prtlen;j++)
		fprintf(f," %10.4e ",
					mysqrt(smatr[i-1][j-1])),smatr1[i-1][j-1]=smatr[i-1][j-1];
	fprintf(f,"\n");
	}
	*/
	if(dia_(&prtlen,(double *)smatr,rmatr,(double *)zmatr))
		return;

	fprintf(f,"\nNumerical values of mixing matrix elements (prev. value):\n\n");
	
/*	for(i=1;i<=prtlen;i++)
	{
	fprintf(f,"s %dx ",i);
	for(j=1;j<=prtlen;j++)
		fprintf(f," %9.2e ",mysqrt(smatr[i-1][j-1]));
	fprintf(f,"\n");
	}
			
	for(i=1;i<=prtlen;i++)
	{
	fprintf(f,"z %dx ",i);
	for(j=1;j<=prtlen;j++)
		fprintf(f," %9.2e ",zmatr[i-1][j-1]);
	fprintf(f,"\n");
	}*/
	
	/*{
		double t1[32][32], t2[32][32];
		int k;
		
		for(i=0;i<32;i++)
			for(j=0;j<32;j++)
				t1[i][j]=t2[i][j]=0.0;
		
		for(i=0;i<prtlen;i++)
			for(j=0;j<prtlen;j++)
				for(k=0;k<prtlen;k++)
					t1[i][j]+=zmatr[i][k]*zmatr[j][k];
		
		
		for(i=1;i<=prtlen;i++)
		{
		fprintf(f,"t1 %dx ",i);
		for(j=1;j<=prtlen;j++)
			fprintf(f," %9.2e ",t1[i-1][j-1]);
		fprintf(f,"\n");
		}
		
		for(i=0;i<32;i++)
			for(j=0;j<32;j++)
				t1[i][j]=t2[i][j]=0.0;
		for(i=0;i<prtlen;i++)
			for(j=0;j<prtlen;j++)
				for(k=0;k<prtlen;k++)
					t1[i][j]+=zmatr[i][k]*smatr1[j][k];
		
		for(i=0;i<prtlen;i++)
			for(j=0;j<prtlen;j++)
				for(k=0;k<prtlen;k++)
					t2[i][j]+=t1[i][k]*zmatr[j][k];
		
		for(i=1;i<=prtlen;i++)
		{
		fprintf(f,"t2 %dx ",i);
		for(j=1;j<=prtlen;j++)
			fprintf(f," %9.2e ",t2[i-1][j-1]);
		fprintf(f,"\n");
		}
	}*/
		
	
	
/*	for(i=1;i<=prtlen;i++)
	for(j=1;j<=prtlen;j++)
		fprintf(f,"s %d %d %f\n",i,j,smatr[i-1][j-1]);
	
	for(i=1;i<=prtlen;i++)
		fprintf(f,"r %d %f\n",i,mysqrt(rmatr[i-1]));*/

	for(i=1;i<=prtlen;i++)
	{
		fprintf(f,"parameter ");
		for(j=1;j<=prtlen;j++)
		{
			Atom v;
			v=herm_matr_el(ola,prtp[i],prtp[j]);
			fprintf(f,"%s=%13g (%13g)",AtomValue(v),zmatr[j-1][i-1],
					EvalParameter(v));
			if(j!=prtlen)
				fprintf(f,", ");
			else
				fprintf(f,".\n");
			if(do_modify)
				ChangeParameterValue(v,zmatr[j-1][i-1]);
		}
	}
		
	fprintf(f,"\nResulting mass matrix:\n\n");
	
	for(i=1;i<=prtlen;i++)
	for(j=1;j<=prtlen;j++)
	{
		Atom p1,p2;
		
		if(!is_c && i>j)
			continue;
		
		p2=prta[i];
		p1=GetAtomProperty(prta[j],A_ANTI);
		if((spi!=1||is_c==0) && (strcmp(AtomValue(p1),AtomValue(p2))>0))
				{
				Atom tmp;
				tmp=p1;
				p1=p2;
				p2=tmp;
				}
				
		fprintf(f,"M%d%d (%s/%s) = ",i,j,AtomValue(p1),AtomValue(p2));
			
		for(l1=al;l1;l1=ListTail(l1))
		{
			if(CompoundArg1(CompoundArg1(ListFirst(l1)))==p1 &&
				CompoundArg2(CompoundArg1(ListFirst(l1)))==p2)
			{
				wrt_num(f,CompoundArgN(ListFirst(l1),5),spi);
				fprintf(f,"\n");
				break;
			}
		}
		if(is_empty_list(l1))
			fprintf(f,"0\n");
	}
		
}

static int correct_kt(Term m2, int spi)
{
	List l1,l2=0;
	Term ga=0;
	if(CompoundArg2(m2))
		return 0;
	if(CompoundArg2(CompoundArg1(m2))!=NewInteger(1))
		return 0;
	if((spi==0||spi==3) && CompoundArg1(CompoundArg1(m2))!=NewInteger(1))
		return 0;
	if((spi!=0&&spi!=3) && CompoundArg1(CompoundArg1(m2))!=NewInteger(-1))
		return 0;
	
	for(l1=CompoundArgN(m2,3);l1;l1=ListTail(l1))
	{
		Term t;
		t=ListFirst(l1);
		if(CompoundName(t)==OPR_SPECIAL && CompoundArg1(t)==A_DELTA)
			continue;
		if(CompoundName(t)==A_MOMENT)
		{
			l2=AppendLast(l2,t);
			continue;
		}
		if(CompoundName(t)==OPR_SPECIAL && CompoundArg1(t)==A_GAMMA && !ga)
		{
			ga=t;
			continue;
		}
		RemoveList(l2);
		return 0;
	}
	

	if(spi==1)
	{
		if(ListLength(l2)!=1 || ga==0 || 
				ListFirst(CompoundArg2(ListFirst(l2)))
				!= ListNth(CompoundArg2(ga),3))
		{
			RemoveList(l2);
			return 0;
		}
		RemoveList(l2);
		return 1;
	}
	
	if(spi==3)
	{
		if(ListLength(l2)==2)
		{
			RemoveList(l2);
			return 1;
		}
		RemoveList(l2);
		return 0;
	}
	
	if(ListLength(l2)!=2 || 
			ListFirst(CompoundArg2(ListFirst(l2)))!=
			ListFirst(CompoundArg2(ListFirst(ListTail(l2)))))
	{
		RemoveList(l2);
		return 0;
	}
	RemoveList(l2);
	return 1;
}


static int correct_ktv(List ml)
{
	if(correct_kt(ListFirst(ml),2) && correct_kt(ListFirst(ListTail(ml)),3))
		return 1;
	if(correct_kt(ListFirst(ml),3) && correct_kt(ListFirst(ListTail(ml)),2))
		return 1;
	return 0;
}
	
static int correct_mt(Term a2, Atom bn, Atom *m)
{
	int spi;
	Term prop;
	
	
	spi=prt_spin(bn);
	prop=GetAtomProperty(bn,PROP_TYPE);
	if(CompoundName(prop)==OPR_FIELD)
	{
		if(CompoundArg2(prop)!=NewInteger(1))
		{
			*m=1;
			return 0;
		}
		prop=GetAtomProperty(CompoundArg1(prop),PROP_TYPE);
	}
	*m=CompoundArgN(prop,5);
	
	
	prop=CompoundArgN(a2,5);
	if(ListLength(prop)!=1)
		return 0;
	
	prop=ListFirst(prop);
	if(CompoundArg2(CompoundArg1(prop))!=NewInteger(1))
		return 0;
	if(CompoundArg1(CompoundArg1(prop))!=NewInteger(1))
		return 0;	
	if(ListLength(CompoundArg2(prop))!=1)
		return 0;
	if(CompoundArg1(ListFirst(CompoundArg2(prop)))!=(*m))
		return 0;
	if(spi==1 && CompoundArg2(ListFirst(CompoundArg2(prop)))!=NewInteger(1))
		return 0;
	if(spi!=1 && CompoundArg2(ListFirst(CompoundArg2(prop)))!=NewInteger(2))
		return 0;
	for(prop=CompoundArgN(prop,3);prop;prop=ListTail(prop))
		if(CompoundName(ListFirst(prop))!=OPR_SPECIAL ||
				CompoundArg1(ListFirst(prop))!=A_DELTA)
			return 0;
	return 1;
	
}


/* dia.f -- translated by f2c (version 19961017).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/


typedef double doublereal;
typedef int integer;


/* Table of constant values */

static integer c__32 = 32;
static /* Subroutine */ int eisch1_();

/*ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
/*    This is just a CERNLIB E202 general complex matrix diagonalization c*/
/*    procedure                                                          c*/
/*ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
static doublereal d_sign(x, y)
doublereal *x, *y;
{
    /* System generated locals */
    doublereal ret_val;

    if (*y > (float)0.) {
	ret_val = *x;
    } else {
	ret_val = -(*x);
    }
    return ret_val;
} /* sign_ */

/* Subroutine */ static int dia_(n, a, am, z__)
integer *n;
doublereal *a, *am, *z__;
{
    static integer ierr, i__, i, j;
    static doublereal w[1024];
    static doublereal ac[1024]	/* was [32][32] */, zc[1024]	/* was [32][
	    32] */;

    /* Parameter adjustments */
    z__ -= 33;
    --am;
    a -= 33;

    /* Function Body */
    for (i__ = 1; i__ <= 32; ++i__) {
	for (j = 1; j <= 32; ++j) {
	    ac[i__ + (j << 5) - 33] = (float)0.;
/* L1: */
	    zc[i__ + (j << 5) - 33] = (float)0.;
	}
    }
    eisch1_(&c__32, n, &a[33], ac, &am[1], &z__[33], zc, &ierr, &w);
    
/*	for(j=1;j<=(*n);j++)
		printf("%f ",mysqrt(am[j]));
	puts("");
	for(i=0;i<(*n);i++)
		{
		printf("c %dx ",i);
		for(j=0;j<(*n);j++)
			printf(" %9.2e ",zc[i*(*n)+j]);
		printf("\n");
		}
	for(j=0;j<(*n);j++)
		printf("%f ",zc[j]);
	puts("");*/
	
	
	return ierr;
} /* dia_ */


static /* Subroutine */ int htridi_(), htribk_(), tql2_();

/* Subroutine */ static int eisch1_(nm, n, ar, ai, wr, zr, zi, ierr, work)
integer *nm, *n;
doublereal *ar, *ai, *wr, *zr, *zi;
integer *ierr;
doublereal *work;
{
    /* System generated locals */
    integer ar_dim1, ar_offset, ai_dim1, ai_offset, zr_dim1, zr_offset, 
	    zi_dim1, zi_offset, i__1, i__2;

    /* Local variables */
    static integer i__, j;

    /* Parameter adjustments */
    zi_dim1 = *nm;
    zi_offset = zi_dim1 + 1;
    zi -= zi_offset;
    zr_dim1 = *nm;
    zr_offset = zr_dim1 + 1;
    zr -= zr_offset;
    ai_dim1 = *nm;
    ai_offset = ai_dim1 + 1;
    ai -= ai_offset;
    ar_dim1 = *nm;
    ar_offset = ar_dim1 + 1;
    ar -= ar_offset;
    --wr;
    --work;

    /* Function Body */
    htridi_(nm, n, &ar[ar_offset], &ai[ai_offset], &wr[1], &zi[zi_offset], &
	    zi[zi_offset], &work[1]);
    i__1 = *n;
    for (i__ = 1; i__ <= i__1; ++i__) {
	i__2 = *n;
	for (j = 1; j <= i__2; ++j) {
/* L50: */
	    zr[i__ + j * zr_dim1] = 0.;
	}
/* L100: */
	zr[i__ + i__ * zr_dim1] = 1.;
    }
    tql2_(nm, n, &wr[1], &zi[zi_offset], &zr[zr_offset], ierr);
    if (*ierr != 0) {
	return 0;
    }
    htribk_(nm, n, &ar[ar_offset], &ai[ai_offset], &work[1], n, &zr[zr_offset]
	    , &zi[zi_offset]);
    return 0;
} /* eisch1_ */

/* Subroutine */ static int htridi_(nm, n, ar, ai, d__, e, e2, tau)
integer *nm, *n;
doublereal *ar, *ai, *d__, *e, *e2, *tau;
{
    /* System generated locals */
    integer ar_dim1, ar_offset, ai_dim1, ai_offset, i__1, i__2, i__3;
    doublereal d__1, d__2;

    /* Builtin functions */
    double sqrt();

    /* Local variables */
    static doublereal f, g, h__;
    static integer i__, j, k, l;
    static doublereal scale, fi, gi, hh;
    static integer ii;
    static doublereal si;
    static integer jp1;

    /* Parameter adjustments */
    tau -= 3;
    --e2;
    --e;
    --d__;
    ai_dim1 = *nm;
    ai_offset = ai_dim1 + 1;
    ai -= ai_offset;
    ar_dim1 = *nm;
    ar_offset = ar_dim1 + 1;
    ar -= ar_offset;

    /* Function Body */
    tau[(*n << 1) + 1] = 1.;
    tau[(*n << 1) + 2] = 0.;
    i__1 = *n;
    for (i__ = 1; i__ <= i__1; ++i__) {
/* L100: */
	d__[i__] = ar[i__ + i__ * ar_dim1];
    }
    i__1 = *n;
    for (ii = 1; ii <= i__1; ++ii) {
	i__ = *n + 1 - ii;
	l = i__ - 1;
	h__ = 0.;
	scale = 0.;
	if (l < 1) {
	    goto L130;
	}
	i__2 = l;
	for (k = 1; k <= i__2; ++k) {
/* L120: */
	    scale = scale + (d__1 = ar[i__ + k * ar_dim1], abs(d__1)) + (d__2 
		    = ai[i__ + k * ai_dim1], abs(d__2));
	}
	if (scale != 0.) {
	    goto L140;
	}
	tau[(l << 1) + 1] = 1.;
	tau[(l << 1) + 2] = 0.;
L130:
	e[i__] = 0.;
	e2[i__] = 0.;
	goto L290;
L140:
	i__2 = l;
	for (k = 1; k <= i__2; ++k) {
	    ar[i__ + k * ar_dim1] /= scale;
	    ai[i__ + k * ai_dim1] /= scale;
	    h__ = h__ + ar[i__ + k * ar_dim1] * ar[i__ + k * ar_dim1] + ai[
		    i__ + k * ai_dim1] * ai[i__ + k * ai_dim1];
/* L150: */
	}
	e2[i__] = scale * scale * h__;
	g = sqrt(h__);
	e[i__] = scale * g;
/* Computing 2nd power */
	d__1 = ar[i__ + l * ar_dim1];
/* Computing 2nd power */
	d__2 = ai[i__ + l * ai_dim1];
	f = sqrt(d__1 * d__1 + d__2 * d__2);
	if (f == 0.) {
	    goto L160;
	}
	tau[(l << 1) + 1] = (ai[i__ + l * ai_dim1] * tau[(i__ << 1) + 2] - ar[
		i__ + l * ar_dim1] * tau[(i__ << 1) + 1]) / f;
	si = (ar[i__ + l * ar_dim1] * tau[(i__ << 1) + 2] + ai[i__ + l * 
		ai_dim1] * tau[(i__ << 1) + 1]) / f;
	h__ += f * g;
	g = g / f + 1.;
	ar[i__ + l * ar_dim1] = g * ar[i__ + l * ar_dim1];
	ai[i__ + l * ai_dim1] = g * ai[i__ + l * ai_dim1];
	if (l == 1) {
	    goto L270;
	}
	goto L170;
L160:
	tau[(l << 1) + 1] = -tau[(i__ << 1) + 1];
	si = tau[(i__ << 1) + 2];
	ar[i__ + l * ar_dim1] = g;
L170:
	f = 0.;
	i__2 = l;
	for (j = 1; j <= i__2; ++j) {
	    g = 0.;
	    gi = 0.;
	    i__3 = j;
	    for (k = 1; k <= i__3; ++k) {
		g = g + ar[j + k * ar_dim1] * ar[i__ + k * ar_dim1] + ai[j + 
			k * ai_dim1] * ai[i__ + k * ai_dim1];
		gi = gi - ar[j + k * ar_dim1] * ai[i__ + k * ai_dim1] + ai[j 
			+ k * ai_dim1] * ar[i__ + k * ar_dim1];
/* L180: */
	    }
	    jp1 = j + 1;
	    if (l < jp1) {
		goto L220;
	    }
	    i__3 = l;
	    for (k = jp1; k <= i__3; ++k) {
		g = g + ar[k + j * ar_dim1] * ar[i__ + k * ar_dim1] - ai[k + 
			j * ai_dim1] * ai[i__ + k * ai_dim1];
		gi = gi - ar[k + j * ar_dim1] * ai[i__ + k * ai_dim1] - ai[k 
			+ j * ai_dim1] * ar[i__ + k * ar_dim1];
/* L200: */
	    }
L220:
	    e[j] = g / h__;
	    tau[(j << 1) + 2] = gi / h__;
	    f = f + e[j] * ar[i__ + j * ar_dim1] - tau[(j << 1) + 2] * ai[i__ 
		    + j * ai_dim1];
/* L240: */
	}
	hh = f / (h__ + h__);
	i__2 = l;
	for (j = 1; j <= i__2; ++j) {
	    f = ar[i__ + j * ar_dim1];
	    g = e[j] - hh * f;
	    e[j] = g;
	    fi = -ai[i__ + j * ai_dim1];
	    gi = tau[(j << 1) + 2] - hh * fi;
	    tau[(j << 1) + 2] = -gi;
	    i__3 = j;
	    for (k = 1; k <= i__3; ++k) {
		ar[j + k * ar_dim1] = ar[j + k * ar_dim1] - f * e[k] - g * ar[
			i__ + k * ar_dim1] + fi * tau[(k << 1) + 2] + gi * ai[
			i__ + k * ai_dim1];
		ai[j + k * ai_dim1] = ai[j + k * ai_dim1] - f * tau[(k << 1) 
			+ 2] - g * ai[i__ + k * ai_dim1] - fi * e[k] - gi * 
			ar[i__ + k * ar_dim1];
/* L260: */
	    }
	}
L270:
	i__3 = l;
	for (k = 1; k <= i__3; ++k) {
	    ar[i__ + k * ar_dim1] = scale * ar[i__ + k * ar_dim1];
	    ai[i__ + k * ai_dim1] = scale * ai[i__ + k * ai_dim1];
/* L280: */
	}
	tau[(l << 1) + 2] = -si;
L290:
	hh = d__[i__];
	d__[i__] = ar[i__ + i__ * ar_dim1];
	ar[i__ + i__ * ar_dim1] = hh;
	ai[i__ + i__ * ai_dim1] = scale * scale * h__;
/* L300: */
    }
    return 0;
} /* htridi_ */

/* Subroutine */ static int htribk_(nm, n, ar, ai, tau, m, zr, zi)
integer *nm, *n;
doublereal *ar, *ai, *tau;
integer *m;
doublereal *zr, *zi;
{
    /* System generated locals */
    integer ar_dim1, ar_offset, ai_dim1, ai_offset, zr_dim1, zr_offset, 
	    zi_dim1, zi_offset, i__1, i__2, i__3;

    /* Local variables */
    static doublereal h__;
    static integer i__, j, k, l;
    static doublereal s, si;

    /* Parameter adjustments */
    tau -= 3;
    ai_dim1 = *nm;
    ai_offset = ai_dim1 + 1;
    ai -= ai_offset;
    ar_dim1 = *nm;
    ar_offset = ar_dim1 + 1;
    ar -= ar_offset;
    zi_dim1 = *nm;
    zi_offset = zi_dim1 + 1;
    zi -= zi_offset;
    zr_dim1 = *nm;
    zr_offset = zr_dim1 + 1;
    zr -= zr_offset;

    /* Function Body */
    i__1 = *n;
    for (k = 1; k <= i__1; ++k) {
	i__2 = *m;
	for (j = 1; j <= i__2; ++j) {
	    zi[k + j * zi_dim1] = -zr[k + j * zr_dim1] * tau[(k << 1) + 2];
	    zr[k + j * zr_dim1] *= tau[(k << 1) + 1];
/* L50: */
	}
    }
    if (*n == 1) {
	goto L200;
    }
    i__2 = *n;
    for (i__ = 2; i__ <= i__2; ++i__) {
	l = i__ - 1;
	h__ = ai[i__ + i__ * ai_dim1];
	if (h__ == 0.) {
	    goto L140;
	}
	i__1 = *m;
	for (j = 1; j <= i__1; ++j) {
	    s = 0.;
	    si = 0.;
	    i__3 = l;
	    for (k = 1; k <= i__3; ++k) {
		s = s + ar[i__ + k * ar_dim1] * zr[k + j * zr_dim1] - ai[i__ 
			+ k * ai_dim1] * zi[k + j * zi_dim1];
		si = si + ar[i__ + k * ar_dim1] * zi[k + j * zi_dim1] + ai[
			i__ + k * ai_dim1] * zr[k + j * zr_dim1];
/* L110: */
	    }
	    s /= h__;
	    si /= h__;
	    i__3 = l;
	    for (k = 1; k <= i__3; ++k) {
		zr[k + j * zr_dim1] = zr[k + j * zr_dim1] - s * ar[i__ + k * 
			ar_dim1] - si * ai[i__ + k * ai_dim1];
		zi[k + j * zi_dim1] = zi[k + j * zi_dim1] - si * ar[i__ + k * 
			ar_dim1] + s * ai[i__ + k * ai_dim1];
/* L120: */
	    }
/* L130: */
	}
L140:
	;
    }
L200:
    return 0;
} /* htribk_ */

/* Subroutine */ static int tql2_(nm, n, d__, e, z__, ierr)
integer *nm, *n;
doublereal *d__, *e, *z__;
integer *ierr;
{
    /* System generated locals */
    integer z_dim1, z_offset, i__1, i__2, i__3;
    doublereal d__1, d__2;

    /* Builtin functions */
    double sqrt(), d_sign();

    /* Local variables */
    static doublereal b, c__, f, g, h__;
    static integer i__, j, k, l, m;
    static doublereal p, r__, s;
    static integer ii;
    static doublereal machep;
    static integer mml;

    /* Parameter adjustments */
    z_dim1 = *nm;
    z_offset = z_dim1 + 1;
    z__ -= z_offset;
    --e;
    --d__;

    /* Function Body */
    machep = 1.1920928955078125e-7;
    *ierr = 0;
    if (*n == 1) {
	goto L1001;
    }
    i__1 = *n;
    for (i__ = 2; i__ <= i__1; ++i__) {
/* L100: */
	e[i__ - 1] = e[i__];
    }
    f = 0.;
    b = 0.;
    e[*n] = 0.;
    i__1 = *n;
    for (l = 1; l <= i__1; ++l) {
	j = 0;
	h__ = machep * ((d__1 = d__[l], abs(d__1)) + (d__2 = e[l], abs(d__2)))
		;
	if (b < h__) {
	    b = h__;
	}
	i__2 = *n;
	for (m = l; m <= i__2; ++m) {
	    if ((d__1 = e[m], abs(d__1)) <= b) {
		goto L120;
	    }
/* L110: */
	}
L120:
	if (m == l) {
	    goto L220;
	}
L130:
	if (j == 30) {
	    goto L1000;
	}
	++j;
	p = (d__[l + 1] - d__[l]) / (e[l] * 2.);
	r__ = sqrt(p * p + 1.);
	h__ = d__[l] - e[l] / (p + d_sign(&r__, &p));
	i__2 = *n;
	for (i__ = l; i__ <= i__2; ++i__) {
/* L140: */
	    d__[i__] -= h__;
	}
	f += h__;
	p = d__[m];
	c__ = 1.;
	s = 0.;
	mml = m - l;
	i__2 = mml;
	for (ii = 1; ii <= i__2; ++ii) {
	    i__ = m - ii;
	    g = c__ * e[i__];
	    h__ = c__ * p;
	    if (abs(p) < (d__1 = e[i__], abs(d__1))) {
		goto L150;
	    }
	    c__ = e[i__] / p;
	    r__ = sqrt(c__ * c__ + 1.);
	    e[i__ + 1] = s * p * r__;
	    s = c__ / r__;
	    c__ = 1. / r__;
	    goto L160;
L150:
	    c__ = p / e[i__];
	    r__ = sqrt(c__ * c__ + 1.);
	    e[i__ + 1] = s * e[i__] * r__;
	    s = 1. / r__;
	    c__ *= s;
L160:
	    p = c__ * d__[i__] - s * g;
	    d__[i__ + 1] = h__ + s * (c__ * g + s * d__[i__]);
	    i__3 = *n;
	    for (k = 1; k <= i__3; ++k) {
		h__ = z__[k + (i__ + 1) * z_dim1];
		z__[k + (i__ + 1) * z_dim1] = s * z__[k + i__ * z_dim1] + c__ 
			* h__;
		z__[k + i__ * z_dim1] = c__ * z__[k + i__ * z_dim1] - s * h__;
/* L180: */
	    }
/* L200: */
	}
	e[l] = s * p;
	d__[l] = c__ * p;
	if ((d__1 = e[l], abs(d__1)) > b) {
	    goto L130;
	}
L220:
	d__[l] += f;
/* L240: */
    }
    i__1 = *n;
    for (ii = 2; ii <= i__1; ++ii) {
	i__ = ii - 1;
	k = i__;
	p = d__[i__];
	i__2 = *n;
	for (j = ii; j <= i__2; ++j) {
	    if (d__[j] >= p) {
		goto L260;
	    }
	    k = j;
	    p = d__[j];
L260:
	    ;
	}
	if (k == i__) {
	    goto L300;
	}
	d__[k] = d__[i__];
	d__[i__] = p;
	i__2 = *n;
	for (j = 1; j <= i__2; ++j) {
	    p = z__[j + i__ * z_dim1];
	    z__[j + i__ * z_dim1] = z__[j + k * z_dim1];
	    z__[j + k * z_dim1] = p;
/* L280: */
	}
L300:
	;
    }
    goto L1001;
L1000:
    *ierr = l;
L1001:
    return 0;
} /* tql2_ */


